/* This file is part of the avgbs2txt project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.avgbs2txt.toTxt;

import ch.ehi.basics.logging.EhiLogger;
import ch.interlis.iom.IomFile;
import ch.interlis.iom.IomObject;
import ch.interlis.iox.*;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import ch.ehi.basics.io.IndentPrintWriter;

/** Implementiert die Konvertierungsfunktion.
 * @author ce
 * @version $Revision: 1.0 $ $Date: 15.12.2006 $
 */
public class TransferToTxt {
	/** Ili-Name des avgbs Modells.
	 */
	private final String ILI_MODEL="GB2AV";
	/** Qualifizierter Ili-Name des Topics Eigentumsverhaeltnis.
	 */
	private final String ILI_EIG=ILI_MODEL+".Eigentumsverhaeltnis";
	/** Qualifizierter Ili-Name des Topics Grundstuecksbeschrieb.
	 */
	private final String ILI_GSBESCHR=ILI_MODEL+".Grundstuecksbeschrieb";
	/** Qualifizierter Ili-Name des Topics Vollzugsgegenstaende.
	 */
	private final String ILI_VZ=ILI_MODEL+".Vollzugsgegenstaende";
	/** Qualifizierter Ili-Name des Topics Mutationstabelle.
	 */
	private final String ILI_MUT=ILI_MODEL+".Mutationstabelle";
	/** Input-Stream.
	 * Wrapper um iomFile.
	 */
	private IoxReader ioxReader=null;
	/** Output-Stream.
	 */
	private IndentPrintWriter out=null;
	/** Hauptfunktion.
	 * Liest aus xmlFile und schreibt in txtFile.
	 * @param xmlFile Name der Eingabedatei (INTERLIS XML gem. avgbs Modell).
	 * @param txtFile Name der Ausgabedatei (TXT).
	 */
	public void doit(String xmlFile,String txtFile)
	{
		// open txt file
		try{
			out=new IndentPrintWriter(new java.io.BufferedWriter(new java.io.FileWriter(txtFile)));
		}catch(java.io.IOException ex){
			EhiLogger.logError(ex);
			throw new IllegalArgumentException("failed to open output file <"+txtFile+">");
		}
		
		try{
			// open xml file
			ioxReader=new ch.interlis.iom_j.xtf.XtfReader(new java.io.File(xmlFile));
			// loop threw baskets
			IoxEvent event;
			while(true){
				event=ioxReader.read();
				//EhiLogger.debug("event "+event.getClass().getName());
				if(event instanceof ObjectEvent){
				}else if(event instanceof StartBasketEvent){
					StartBasketEvent se=(StartBasketEvent)event;
					String topic=se.getType();
					if(topic.equals(ILI_EIG)){
						doEigentumsverhaeltnis(se);
					}else if(topic.equals(ILI_GSBESCHR)){
						doGrundstuecksbeschrieb(se);
					}else if(topic.equals(ILI_VZ)){
						doVollzugsgegenstaende(se);
					}else if(topic.equals(ILI_MUT)){
						doMutationstabelle(se);
					}else{
						EhiLogger.logAdaption("unexpected basket "+topic+"; ignored");
					}
				}else if(event instanceof EndBasketEvent){
				}else if(event instanceof EndTransferEvent){
					out.flush();
					ioxReader.close();
					ioxReader=null;
					break;
				}else{
					// ignore other events
				}
			}
		}catch(IoxException ex){
			EhiLogger.logError(ex);
		}finally{
			if(out!=null){
				out.close();
				out=null;
			}
			if(ioxReader!=null){
				try{
					ioxReader.close();
				}catch(IoxException ex){
					EhiLogger.logError(ex);
				}
				ioxReader=null;
			}
			
		}
	}
	/** liest und konvertiert ein Basket gemaess Topic Eigentumsverhaeltnis.
	 * @param basket Header des Baskets.
	 */
	private void doEigentumsverhaeltnis(StartBasketEvent basket)
	{
		out.println(basket.getType()+" (BID "+basket.getBid()+")");
		// alle Objekte dieses Baskets
		HashMap objv=new HashMap();
		// alle Eigentumsobjekte pro Grundstueck-OID
		HashMap eigentumv=new HashMap(); // map<String gsOid,ArrayList<IomObject eigentum>>
		// alle GemeinschaftTeilhaber-Objekte pro Gemeinschaft-OID
		HashMap gemeinschaftTeilhaberv=new HashMap(); // map<String gemeinschaftOid,ArrayList<IomObject GemeinschaftTeilhaber>>
		// alle Grundstuecke
		ArrayList gsv=new ArrayList();
		try{
			
			// loop threw basket
			IoxEvent event;
			while(true){
				event=ioxReader.read();
				//EhiLogger.debug("event "+event.getClass().getName());
				if(event instanceof ObjectEvent){
					IomObject iomObj=((ObjectEvent)event).getIomObject();
					if(isOidOk(objv,iomObj)){
						objv.put(iomObj.getobjectoid(),iomObj);
						String aclass=iomObj.getobjecttag();
						if(aclass.equals(ILI_EIG+".Grundstueck")){
							gsv.add(iomObj);
						}
						collectLinks(eigentumv,ILI_EIG+".RechtGrundstueck","betroffenesGrundstueck",iomObj);
						collectLinks(eigentumv,ILI_EIG+".RechtPerson","betroffenesGrundstueck",iomObj);
						collectLinks(gemeinschaftTeilhaberv,ILI_EIG+".GemeinschaftTeilhaber","Gemeinschaft",iomObj);
					}
				}else if(event instanceof EndBasketEvent){
					break;
				}else{
					throw new IllegalStateException("unexpected event "+event.getClass().getName());
				}
			}
		}catch(IoxException ex){
			EhiLogger.logError(ex);
		}
		out.indent();
		for(int gsi=0;gsi<gsv.size();gsi++){
			IomObject iomObj=(IomObject)gsv.get(gsi);
			IomObject gsNr=iomObj.getattrobj("Nummer",0);
			String gsNrTxt=formatGSNr(gsNr,iomObj.getobjectoid());
			out.println();
			out.println("Grundstck "+gsNrTxt);	
			out.indent();
			String Art=iomObj.getattrvalue("Art");
			if(Art!=null){
				out.println("Art "+Art);
			}
			if(!eigentumv.containsKey(iomObj.getobjectoid())){
				EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": keine Eigentumsobjekte vorhanden");
			}else{
				ArrayList eigv=(ArrayList)eigentumv.get(iomObj.getobjectoid());
				for(int eigi=0;eigi<eigv.size();eigi++){
					IomObject eig=(IomObject)eigv.get(eigi);
					String aclass=eig.getobjecttag();
					StringBuffer txt=new StringBuffer();
					String sep="Eigentum ";
					String AnteilZaehler=eig.getattrvalue("AnteilZaehler");
					String AnteilNenner=eig.getattrvalue("AnteilNenner");
					if(AnteilZaehler!=null && AnteilNenner!=null){
						txt.append(sep+AnteilZaehler+"/"+AnteilNenner);sep=", ";
					}else{
						if(AnteilZaehler!=null){
							txt.append(sep+"AnteilZaehler "+AnteilZaehler);sep=", ";
						}
						if(AnteilNenner!=null){
							txt.append(sep+"AnteilNenner "+AnteilNenner);sep=", ";
						}
					}
					String Bemerkungen=eig.getattrvalue("Bemerkungen");
					if(Bemerkungen!=null){
						txt.append(sep+"Bemerkungen "+Bemerkungen);sep=", ";
					}
					out.println(txt);
					out.indent();
					if(aclass.equals(ILI_EIG+".RechtGrundstueck")){
						String SubjektivDinglich=eig.getattrvalue("SubjektivDinglich");
						if(SubjektivDinglich!=null){
							out.println("SubjektivDinglich "+SubjektivDinglich);
						}
						IomObject refObj=eig.getattrobj("berechtigtesGrundstueck",0);
						if(refObj!=null){
							String gsOid=refObj.getobjectrefoid();
							if(!objv.containsKey(gsOid)){
								EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": enthlt Verweis auf nicht existierendes Objekt ("+gsOid+")");
							}else{
								IomObject eigGs=(IomObject)objv.get(gsOid);
								IomObject eigGsNr=eigGs.getattrobj("Nummer",0);
								String eigGsNrTxt=formatGSNr(eigGsNr,gsOid);
								out.println("berechtigtes Grundstck "+eigGsNrTxt);	
							}
						}else{
							EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": kein Verweis berechtigtesGrundstueck vorhanden");
						}
					}else if(aclass.equals(ILI_EIG+".RechtPerson")){
						String EigArt=eig.getattrvalue("Art");
						if(EigArt!=null){
							out.println("Art "+EigArt);
						}
						IomObject refObj=eig.getattrobj("Berechtigter",0);
						if(refObj!=null){
							String personOid=refObj.getobjectrefoid();
							if(!objv.containsKey(personOid)){
								EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": enthlt Verweis auf nicht existierendes Objekt ("+personOid+")");
							}else{
								IomObject person=(IomObject)objv.get(personOid);
								printPerson(person,objv,gemeinschaftTeilhaberv);
							}
						}
					}else{
						throw new IllegalStateException("unexpected object "+aclass);
					}
					out.unindent();
				}
			}
			out.unindent();
		}
		out.unindent();
	}
	/** sammelt Assoziationsobjekte und gruppiert sie nach OIDs eines Assoziationsendes.
	 * @param allLinks alle gesammelten Assoziationsobjekte map<String oid,list<IomObject link>>
	 * @param assocName Qualifizierter Ili-Name der Assoziation.
	 * @param parentRoleName Rollenname (Assoziationsende) nach dem die Links gruppiert werden sollen.
	 * @param iomObj fuer die Sammlung in Frgae kommendes Objekt
	 */
	private void collectLinks(HashMap allLinks,String assocName,String parentRoleName,IomObject iomObj)
	{
		String aclass=iomObj.getobjecttag();
		if(aclass.equals(assocName)){
			IomObject refObj=iomObj.getattrobj(parentRoleName,0);
			if(refObj!=null){
				String parentOid=refObj.getobjectrefoid();
				ArrayList linkv=null;
				if(allLinks.containsKey(parentOid)){
					linkv=(ArrayList)allLinks.get(parentOid);
				}else{
					linkv=new ArrayList();
					allLinks.put(parentOid,linkv);
				}
				linkv.add(iomObj);
			}
		}
	}
	/** liest und konvertiert ein Basket gemaess Topic Grundstuecksbeschrieb.
	 * @param basket Header des Baskets.
	 */
	private void doGrundstuecksbeschrieb(StartBasketEvent basket)
	{
		out.println(basket.getType()+" (BID "+basket.getBid()+")");
		HashMap objv=new HashMap();
		try{
			
			// loop threw basket
			IoxEvent event;
			while(true){
				event=ioxReader.read();
				//EhiLogger.debug("event "+event.getClass().getName());
				if(event instanceof ObjectEvent){
					IomObject iomObj=((ObjectEvent)event).getIomObject();
					if(isOidOk(objv,iomObj)){
						objv.put(iomObj.getobjectoid(),iomObj);
					}
				}else if(event instanceof EndBasketEvent){
					break;
				}else{
					throw new IllegalStateException("unexpected event "+event.getClass().getName());
				}
			}
		}catch(IoxException ex){
			EhiLogger.logError(ex);
		}
		out.indent();
		printGSBeschrieb(objv,ILI_GSBESCHR);			
		out.unindent();
	}
	/** liest und konvertiert ein Basket gemaess Topic Vollzugsgegenstaende.
	 * @param basket Header des Baskets.
	 */
	private void doVollzugsgegenstaende(StartBasketEvent basket)
	{
		out.println(basket.getType()+" (BID "+basket.getBid()+")");
		HashMap objv=new HashMap();
		ArrayList mutv=new ArrayList();
		try{
			
			// loop threw basket
			IoxEvent event;
			while(true){
				event=ioxReader.read();
				//EhiLogger.debug("event "+event.getClass().getName());
				if(event instanceof ObjectEvent){
					IomObject iomObj=((ObjectEvent)event).getIomObject();
					if(isOidOk(objv,iomObj)){
						objv.put(iomObj.getobjectoid(),iomObj);
						String aclass=iomObj.getobjecttag();
						if(aclass.equals(ILI_VZ+".Vollzugsgegenstand")){
							mutv.add(iomObj);
						}
					}
				}else if(event instanceof EndBasketEvent){
					break;
				}else{
					throw new IllegalStateException("unexpected event "+event.getClass().getName());
				}
			}
		}catch(IoxException ex){
			EhiLogger.logError(ex);
		}
		if(mutv.size()==0){
			EhiLogger.logError("kein Vollzugsgegenstand vorhanden");
			return;
		}
		out.indent();
		Iterator muti=mutv.iterator();
		while(muti.hasNext()){
			IomObject mut=(IomObject)muti.next();
			IomObject mutnr=mut.getattrobj("Mutationsnummer",0);
			out.println("Vollzugsgegenstand NBIdent "+mutnr.getattrvalue("NBIdent")+", Nummer "+mutnr.getattrvalue("Nummer"));				
			out.indent();
			String Status=mut.getattrvalue("Status");
			if(Status!=null){
				out.println("Status "+Status);
			}
			String Bemerkungen=mut.getattrvalue("Bemerkungen");
			if(Bemerkungen!=null){
				out.println("Bemerkungen "+Bemerkungen);
			}
			String GrundbuchEintrag=mut.getattrvalue("GrundbuchEintrag");
			if(GrundbuchEintrag!=null){
				out.println("GrundbuchEintrag "+GrundbuchEintrag);
			}
			String TagebuchEintrag=mut.getattrvalue("TagebuchEintrag");
			if(TagebuchEintrag!=null){
				out.println("TagebuchEintrag "+TagebuchEintrag);
			}
			String TagebuchBeleg=mut.getattrvalue("TagebuchBeleg");
			if(TagebuchBeleg!=null){
				out.println("TagebuchBeleg "+TagebuchBeleg);
			}
		}
		out.unindent();
	}
	/** liest und konvertiert ein Basket gemaess Topic Mutationstabelle.
	 * @param basket Header des Baskets.
	 */
	private void doMutationstabelle(StartBasketEvent basket)
	{
		out.println(basket.getType()+" (BID "+basket.getBid()+")");
		HashMap objv=new HashMap();
		ArrayList mutv=new ArrayList();
		ArrayList mutannulationv=new ArrayList();
		ArrayList mutBetrfGs=new ArrayList();
		ArrayList anteilv=new ArrayList();

		try{
			
			// loop threw basket
			IoxEvent event;
			while(true){
				event=ioxReader.read();
				//EhiLogger.debug("event "+event.getClass().getName());
				if(event instanceof ObjectEvent){
					IomObject iomObj=((ObjectEvent)event).getIomObject();
					if(isOidOk(objv,iomObj)){
						objv.put(iomObj.getobjectoid(),iomObj);
						String aclass=iomObj.getobjecttag();
						//EhiLogger.debug("object "+aclass);
						if(aclass.equals(ILI_MUT+".AVMutation")){
							mutv.add(iomObj);
						}
						if(aclass.equals(ILI_MUT+".AVMutationsAnnulation")){
							mutannulationv.add(iomObj);
						}
						if(aclass.equals(ILI_MUT+".AVMutationBetroffeneGrundstuecke")){
							//EhiLogger.debug(iomObj.toString());
							mutBetrfGs.add(iomObj);
						}
						if(aclass.equals(ILI_GSBESCHR+".Anteil")){
							anteilv.add(iomObj);
						}
					}
				}else if(event instanceof EndBasketEvent){
					break;
				}else{
					throw new IllegalStateException("unexpected event "+event.getClass().getName());
				}
			}
		}catch(IoxException ex){
			EhiLogger.logError(ex);
		}
		
			if(mutannulationv.size()==0 && mutv.size()==0){
				EhiLogger.logError("keine Mutation und keine Annulation vorhanden");
				return;
			}
			
			out.indent();
			// Annulationen ausgeben
			Iterator mutannulationi=mutannulationv.iterator();
			while(mutannulationi.hasNext()){
				IomObject annulation=(IomObject)mutannulationi.next();
				IomObject mutnr=annulation.getattrobj("MutationsNummer",0);
				out.println("Annulation der Mutation NBIdent "+mutnr.getattrvalue("NBIdent")+", Nummer "+mutnr.getattrvalue("Nummer"));				
				String bem=annulation.getattrvalue("Bemerkungen");
				if(bem!=null){
					out.indent();
					out.println(bem);
					out.unindent();
				}
			}
			
			// Mutationen ausgeben
			Iterator muti=mutv.iterator();
			while(muti.hasNext()){
				IomObject mut=(IomObject)muti.next();
				IomObject mutnr=mut.getattrobj("MutationsNummer",0);
				out.println("Mutation NBIdent "+mutnr.getattrvalue("NBIdent")+", Nummer "+mutnr.getattrvalue("Nummer"));				
				out.indent();
				String beschrieb=mut.getattrvalue("Beschrieb");
				if(beschrieb!=null){
					out.println("Beschrieb "+beschrieb);
				}
				String dateinamePlan=mut.getattrvalue("DateinamePlan");
				if(dateinamePlan!=null){
					out.println("DateinamePlan "+dateinamePlan);
				}
				String endeTechnBereit=mut.getattrvalue("EndeTechnBereit");
				if(endeTechnBereit!=null){
					out.println("EndeTechnBereit "+endeTechnBereit);
				}
				String istProjektMutation=mut.getattrvalue("istProjektMutation");
				if(istProjektMutation!=null){
					out.println("istProjektMutation "+istProjektMutation);
				}
				int gelGSc=mut.getattrvaluecount("geloeschteGrundstuecke");
				if(gelGSc>0){
					out.println("Gelschte Grundstcke");
					out.indent();
					for(int gelGSi=0;gelGSi<gelGSc;gelGSi++){
						IomObject gsNr=mut.getattrobj("geloeschteGrundstuecke",gelGSi);
						String gsNrTxt=formatGSNr(gsNr,null);
						out.println("Grundstck "+gsNrTxt);					
					}
					out.unindent();
				}
				// betroffene Grundstcke
				ArrayList betrfGSv=getBetrfGS(mut.getobjectoid(),objv,mutBetrfGs);
				int betrfGSc=betrfGSv.size();
				if(betrfGSc>0){
					out.println("Betroffene Grundstcke");
					out.indent();
					for(int betrfGSi=0;betrfGSi<betrfGSc;betrfGSi++){
						IomObject betrfGS=(IomObject)betrfGSv.get(betrfGSi);
						IomObject gsNr=betrfGS.getattrobj("Nummer",0);
						String gsNrTxt=formatGSNr(gsNr,betrfGS.getobjectoid());
						out.println();	
						out.println("Grundstck "+gsNrTxt);	
						out.indent();
						String GrundstueckArt=betrfGS.getattrvalue("GrundstueckArt");
						if(GrundstueckArt!=null){
							out.println("GrundstueckArt "+GrundstueckArt);
						}
						String KantonaleUnterArt=betrfGS.getattrvalue("KantonaleUnterArt");
						if(KantonaleUnterArt!=null){
							out.println("KantonaleUnterArt "+KantonaleUnterArt);
						}
						String Flaechenmass=betrfGS.getattrvalue("Flaechenmass");
						if(Flaechenmass!=null){
							out.println("Flaechenmass "+Flaechenmass);
						}
						String Korrektur=betrfGS.getattrvalue("Korrektur");
						if(Korrektur!=null){
							out.println("Korrektur "+Korrektur);
						}
						String korrigiertesFlaechenmass=betrfGS.getattrvalue("korrigiertesFlaechenmass");
						if(korrigiertesFlaechenmass!=null){
							out.println("korrigiertesFlaechenmass "+korrigiertesFlaechenmass);
						}
						if((ILI_MUT+".Liegenschaft").equals(betrfGS.getobjecttag())){
							int Zugangc=betrfGS.getattrvaluecount("Zugang");
							out.println("Zugang");
							out.indent();
							for(int Zugangi=0;Zugangi<Zugangc;Zugangi++){
								IomObject Zugang=betrfGS.getattrobj("Zugang",Zugangi);
								IomObject vonGsNr=Zugang.getattrobj("von",0);
								String vonGsNrTxt=formatGSNr(vonGsNr,null);
								out.println("Flaechenmass "+Zugang.getattrvalue("Flaechenmass")+" von Grundstck "+vonGsNrTxt);					
							}
							out.unindent();
						}else if((ILI_MUT+".Flaeche").equals(betrfGS.getobjecttag())){
							ArrayList liegtAuf=getLiegtAufGS(betrfGS.getobjectoid(),objv,anteilv);
							int liegtAufc=liegtAuf.size();
							out.println("liegt auf");
							out.indent();
							for(int liegtAufi=0;liegtAufi<liegtAufc;liegtAufi++){
								IomObject Anteil=(IomObject)liegtAuf.get(liegtAufi);
								String liegtAufGsOid=Anteil.getattrobj("liegt_auf",0).getobjectrefoid();
								IomObject liegtAufGs=(IomObject)objv.get(liegtAufGsOid);
								IomObject liegtAufGsNr=liegtAufGs.getattrobj("Nummer",0);
								String liegtAufGsNrTxt=formatGSNr(liegtAufGsNr,liegtAufGsOid);
								String fm=Anteil.getattrvalue("Flaechenmass");
								if(fm!=null){
									out.println("Flaechenmass "+fm+" liegt auf Grundstck "+liegtAufGsNrTxt);					
								}else{
									out.println("Liegt auf Grundstck "+liegtAufGsNrTxt);					
								}
							}
							out.unindent();
						}
						out.unindent();				
					}
					out.unindent();
				}
				
				out.unindent();
			}
			
			printGSBeschrieb(objv,ILI_MUT);
			
			out.unindent();
	}
	/** pueft ob die OID eines gegebenen Objektes korrekt ist.
	 * Gibt eine Fehlermeldung aus.
	 * @param objv Alle schon gelesenen Objekte.
	 * @param iomObj Objekt dessen OID geprueft werden soll.
	 * @return true wenn OID ok ist, sonst false.
	 */
	private boolean isOidOk(HashMap objv,IomObject iomObj)
	{
		String oid=iomObj.getobjectoid();
		if(oid!=null && oid.length()>0){
			if(!objv.containsKey(oid)){
				return true;
			}else{
				EhiLogger.logError("Objekt "+iomObj.getobjecttag()+", Zeile "+iomObj.getobjectline()+": hat doppelte TID "+oid+"; Objekt ignoriert");
			}
		}else{
			EhiLogger.logError("Objekt "+iomObj.getobjecttag()+", Zeile "+iomObj.getobjectline()+": hat keine TID; Objekt ignoriert");
		}
		return false;
	}
	private void printGSBeschrieb(HashMap objv,String topic) {
		ArrayList avbemGSv=new ArrayList();
		HashMap grundstueckGebaeudev=new HashMap();
		// GrundstueckGebaeude
		Iterator oidi=objv.keySet().iterator();
		while(oidi.hasNext()){
			IomObject iomObj=(IomObject)objv.get(oidi.next());
			collectLinks(grundstueckGebaeudev,ILI_GSBESCHR+".GrundstueckGebaeude","Grundstueck",iomObj);
		}

		// Grundstcksbeschrieb
		out.println("Grundstcksbeschrieb");
		out.indent();
		oidi=objv.keySet().iterator();
		while(oidi.hasNext()){
			IomObject iomObj=(IomObject)objv.get(oidi.next());
			String aclass=iomObj.getobjecttag();
			// Ist es ein Grundstck?
			if(aclass.equals(topic+".Liegenschaft") || aclass.equals(topic+".Flaeche")){
				IomObject gsNr=iomObj.getattrobj("Nummer",0);
				String gsNrTxt=formatGSNr(gsNr,iomObj.getobjectoid());
				out.println();
				out.println("Grundstck "+gsNrTxt);	
				out.indent();
				printGS(iomObj);
				if(grundstueckGebaeudev.containsKey(iomObj.getobjectoid())){
					ArrayList gs2gebv=(ArrayList)grundstueckGebaeudev.get(iomObj.getobjectoid());
					for(int gs2gebi=0;gs2gebi<gs2gebv.size();gs2gebi++){
						IomObject gs2geb=(IomObject)gs2gebv.get(gs2gebi);
						IomObject gebRef=gs2geb.getattrobj("Gebaeude",0);
						if(gebRef!=null){
							String gebOid=gebRef.getobjectrefoid();
							if(!objv.containsKey(gebOid)){
							}else{
								IomObject geb=(IomObject)objv.get(gebOid);
								String anteil=null;
								anteil=gs2geb.getattrprim("Flaechenmass",0);
								printGebaeude(geb,null,anteil);
							}
						}
					}
				}
				out.unindent();
			}
			if(aclass.equals(ILI_GSBESCHR+".AVBemerkungGrundstueck")){
				avbemGSv.add(iomObj);
			}
		}
		out.unindent();
		
		// Gebaeude
		if(false){
			oidi=objv.keySet().iterator();
			while(oidi.hasNext()){
				IomObject gebaeude=(IomObject)objv.get(oidi.next());
				String aclass=gebaeude.getobjecttag();
				// Ist es ein Gebaeude?
				if(aclass.equals(ILI_GSBESCHR+".Gebaeude")){
					printGebaeude(gebaeude,objv,null);
				}
			}
		}
		
		// AVBemerkungen
		oidi=objv.keySet().iterator();
		while(oidi.hasNext()){
			IomObject avbem=(IomObject)objv.get(oidi.next());
			String aclass=avbem.getobjecttag();
			// Ist es eine AVBemerkung?
			if(aclass.equals(ILI_GSBESCHR+".AVBemerkung")){
				printAVBemerkung(avbem,objv,avbemGSv);
			}
		}
	}
	/** konvertiert eine AVBemerkung.
	 * @param avbem AVBemerkung Objekt das konvertiert werden soll.
	 * @param objv alle Objekte.
	 * @param avbemGSv Liste aller AVBemerkungGrundstueck Objekte list<IomObject>
	 */
	private void printAVBemerkung(IomObject avbem,HashMap objv,ArrayList avbemGSv)
	{
		StringBuffer txt=new StringBuffer();
		txt.append("AV-Bemerkung");
		String sep=" ";
		String Art=avbem.getattrvalue("Art");
		if(Art!=null){
			txt.append(sep+Art);sep=", ";
		}
		String AndereArt=avbem.getattrvalue("AndereArt");
		if(AndereArt!=null){
			txt.append(sep+"AndereArt "+AndereArt);sep=", ";
		}
		String Bemerkung=avbem.getattrvalue("Bemerkung");
		if(Bemerkung!=null){
			txt.append(sep+"Bemerkung "+Bemerkung);sep=", ";
		}
		txt.append(" (TID "+avbem.getobjectoid()+")");
		out.println(txt);
		out.indent();
		String bemOid=avbem.getobjectoid();
		// Grundstcke auf dieser AVBemerkung
		Iterator oidi=avbemGSv.iterator();
		while(oidi.hasNext()){
			IomObject iomObj=(IomObject)oidi.next();
			String aclass=iomObj.getobjecttag();
			// Ist es ein Assoziationsobjekt?
			if(aclass.equals(ILI_GSBESCHR+".AVBemerkungGrundstueck")){
				IomObject refObj=iomObj.getattrobj("AVBemerkung",0);
				if(refObj!=null){
					// Ist das eine Assoziation mit dieser AVBemerkung?
					if(bemOid.equals(refObj.getobjectrefoid())){
						IomObject refGS=iomObj.getattrobj("Grundstueck",0);
						String gsOid=refGS.getobjectrefoid();
						// check that oid exists
						if(objv.containsKey(gsOid)){
							IomObject gs=(IomObject)objv.get(gsOid);
							IomObject gsNr=gs.getattrobj("Nummer",0);
							String gsNrTxt=formatGSNr(gsNr,gsOid);
							out.println("liegt auf Grundstck "+gsNrTxt);	
						}else{
							EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": enthlt Verweis auf nicht existierendes Objekt ("+gsOid+")");
						}
					}
				}
			}
		}
		out.unindent();
	}
	/** konvertiert ein Gebaeude.
	 * @param gebaeude Gebaeude Objekt das konvertiert wird.
	 * @param objv Alle Objekte.
	 * @param Anteil2 null oder Flaechenmass des Anteils des Gebaeudes auf dem Grundsteuck (z.B. "50")
	 */
	private void printGebaeude(IomObject gebaeude,HashMap objv,String Anteil2)
	{
		String gebNrTxt=formatGebaeudeNr(gebaeude);
		out.println("Gebude "+gebNrTxt);	
		out.indent();

		String Flaechenmass=gebaeude.getattrvalue("Flaechenmass");
		if(Flaechenmass!=null){
			if(Anteil2!=null){
				out.println("Flaechenmass "+Flaechenmass+" ("+Anteil2+" auf diesem Grundstck)");
			}else{
				out.println("Flaechenmass "+Flaechenmass);
			}
		}
		StringBuffer txt=new StringBuffer();
		String sep="";
		String istProjektiert=gebaeude.getattrvalue("istProjektiert");
		if(istProjektiert!=null){
			txt.append(sep+"istProjektiert "+istProjektiert);sep=", ";
		}
		String istUnterirdisch=gebaeude.getattrvalue("istUnterirdisch");
		if(istUnterirdisch!=null){
			txt.append(sep+"istUnterirdisch "+istUnterirdisch);sep=", ";
		}
		String Gebaeudeart=gebaeude.getattrvalue("Gebaeudeart");
		if(Gebaeudeart!=null){
			txt.append(sep+"Gebaeudeart "+Gebaeudeart);sep=", ";
		}
		String GebaeudeartCode=gebaeude.getattrvalue("GebaeudeartCode");
		if(GebaeudeartCode!=null){
			txt.append(sep+"GebaeudeartCode "+GebaeudeartCode);sep=", ";
		}
		out.println(txt);
		out.indent();
		String gebOid=gebaeude.getobjectoid();
		if(objv!=null){
			// Grundstcke auf diesem Gebaeude
			Iterator oidi=objv.keySet().iterator();
			while(oidi.hasNext()){
				IomObject iomObj=(IomObject)objv.get(oidi.next());
				String aclass=iomObj.getobjecttag();
				// Ist es ein Assoziationsobjekt?
				if(aclass.equals(ILI_GSBESCHR+".GrundstueckGebaeude")){
					IomObject refObj=iomObj.getattrobj("Gebaeude",0);
					if(refObj!=null){
						// Ist das eine Assoziation mit diesem Gebaeude?
						if(gebOid.equals(refObj.getobjectrefoid())){
							IomObject refGS=iomObj.getattrobj("Grundstueck",0);
							String gsOid=refGS.getobjectrefoid();
							if(objv.containsKey(gsOid)){
								IomObject gs=(IomObject)objv.get(gsOid);
								IomObject gsNr=gs.getattrobj("Nummer",0);
								String gsNrTxt=formatGSNr(gsNr,gsOid);
								String Anteil=iomObj.getattrvalue("Flaechenmass");
								if(Anteil!=null){
									out.println("Flchenmass "+Anteil+" liegt auf Grundstck "+gsNrTxt);	
								}else{
									out.println("liegt auf Grundstck "+gsNrTxt);	
								}
							}else{
								EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": enthlt Verweis auf nicht existierendes Objekt ("+gsOid+")");
							}
						}
					}
				}
			}
		}
		// Adresse
		int adrc=gebaeude.getattrvaluecount("Adresse");
		for(int adri=0;adri<adrc;adri++){
			IomObject adr=gebaeude.getattrobj("Adresse",adri);
			txt=new StringBuffer();
			txt.append("Adresse");
			sep=" ";
			String Strasse=adr.getattrvalue("Strasse");
			if(Strasse!=null){
				txt.append(sep+"Strasse "+Strasse);sep=", ";
			}
			String Hausnummer=adr.getattrvalue("Hausnummer");
			if(Hausnummer!=null){
				txt.append(sep+"Hausnummer "+Hausnummer);sep=", ";
			}
			String GebaeudeBezeichnung=adr.getattrvalue("GebaeudeBezeichnung");
			if(GebaeudeBezeichnung!=null){
				txt.append(sep+"GebaeudeBezeichnung "+GebaeudeBezeichnung);sep=", ";
			}
			String BenanntesGebiet=adr.getattrvalue("BenanntesGebiet");
			if(BenanntesGebiet!=null){
				txt.append(sep+"BenanntesGebiet "+BenanntesGebiet);sep=", ";
			}
			String PLZ=adr.getattrvalue("PLZ");
			if(PLZ!=null){
				txt.append(sep+"PLZ "+PLZ);sep=", ";
			}
			String PLZZusatzziffern=adr.getattrvalue("PLZZusatzziffern");
			if(PLZZusatzziffern!=null){
				txt.append(sep+"PLZZusatzziffern "+PLZZusatzziffern);sep=", ";
			}
			String Ortschaft=adr.getattrvalue("Ortschaft");
			if(Ortschaft!=null){
				txt.append(sep+"Ortschaft "+Ortschaft);sep=", ";
			}
			String GWR_EGID=adr.getattrvalue("GWR_EGID");
			if(GWR_EGID!=null){
				txt.append(sep+"GWR_EGID "+GWR_EGID);sep=", ";
			}
			String GWR_EDID=adr.getattrvalue("GWR_EDID");
			if(GWR_EDID!=null){
				txt.append(sep+"GWR_EDID "+GWR_EDID);sep=", ";
			}
			out.println(txt);
		}
		out.unindent();
		
		out.unindent();
	}
	/** konvertiert eine Person.
	 * @param person Person Objekt das konvertiert werden soll.
	 * @param objv alle Objekte des Baskets.
	 * @param gemeinschaftTeilhaberv alle GemeinschaftTeilhaber Objekte gruppiert nach Gemeinschaft. map<String oid,list<IomObject link>>
	 */
	private void printPerson(IomObject person,HashMap objv,HashMap gemeinschaftTeilhaberv)
	{
		out.println("Person");
		out.indent();
		String Art=person.getattrvalue("Art");
		if(Art!=null){
			out.println("Art "+Art);
		}
		StringBuffer txt=new StringBuffer();
		String sep="";
		String Name=person.getattrvalue("Name");
		if(Name!=null){
			txt.append(sep+"Name "+Name);sep=", ";
		}
		String Vorname=person.getattrvalue("Vorname");
		if(Vorname!=null){
			txt.append(sep+"Vorname "+Vorname);sep=", ";
		}
		String Geburtsjahr=person.getattrvalue("Geburtsjahr");
		if(Geburtsjahr!=null){
			txt.append(sep+"Geburtsjahr "+Geburtsjahr);sep=", ";
		}
		String Geburtsmonat=person.getattrvalue("Geburtsmonat");
		if(Geburtsmonat!=null){
			txt.append(sep+"Geburtsmonat "+Geburtsmonat);sep=", ";
		}
		String Geburtstag=person.getattrvalue("Geburtstag");
		if(Geburtstag!=null){
			txt.append(sep+"Geburtstag "+Geburtstag);sep=", ";
		}
		String Geschlecht=person.getattrvalue("Geschlecht");
		if(Geschlecht!=null){
			txt.append(sep+"Geschlecht "+Geschlecht);sep=", ";
		}
		String Sitz=person.getattrvalue("Sitz");
		if(Sitz!=null){
			txt.append(sep+"Sitz "+Sitz);sep=", ";
		}
		String EPID=person.getattrvalue("EPID");
		if(EPID!=null){
			txt.append(sep+"EPID "+EPID);sep=", ";
		}
		String UID=person.getattrvalue("UID");
		if(UID!=null){
			txt.append(sep+"UID "+UID);sep=", ";
		}
		out.println(txt);
		
		
		txt=new StringBuffer();
		sep="";
		String Strasse=person.getattrvalue("Strasse");
		if(Strasse!=null){
			txt.append(sep+"Strasse "+Strasse);sep=", ";
		}
		String Hausnummer=person.getattrvalue("Hausnummer");
		if(Hausnummer!=null){
			txt.append(sep+"Hausnummer "+Hausnummer);sep=", ";
		}
		String GebaeudeBezeichnung=person.getattrvalue("GebaeudeBezeichnung");
		if(GebaeudeBezeichnung!=null){
			txt.append(sep+"GebaeudeBezeichnung "+GebaeudeBezeichnung);sep=", ";
		}
		String Ortsteil=person.getattrvalue("Ortsteil");
		if(Ortsteil!=null){
			txt.append(sep+"Ortsteil "+Ortsteil);sep=", ";
		}
		String PLZ=person.getattrvalue("PLZ");
		if(PLZ!=null){
			txt.append(sep+"PLZ "+PLZ);sep=", ";
		}
		String Ortschaft=person.getattrvalue("Ortschaft");
		if(Ortschaft!=null){
			txt.append(sep+"Ortschaft "+Ortschaft);sep=", ";
		}
		String Land=person.getattrvalue("Land");
		if(Land!=null){
			txt.append(sep+"Land "+Land);sep=", ";
		}
		String ZusatzinformationEmpfaenger=person.getattrvalue("ZusatzinformationEmpfaenger");
		if(ZusatzinformationEmpfaenger!=null){
			txt.append(sep+"ZusatzinformationEmpfaenger "+ZusatzinformationEmpfaenger);sep=", ";
		}
		String PostDienstleistung=person.getattrvalue("PostDienstleistung");
		if(PostDienstleistung!=null){
			txt.append(sep+"PostDienstleistung "+PostDienstleistung);sep=", ";
		}
		out.println(txt);
		
		if(gemeinschaftTeilhaberv.containsKey(person.getobjectoid())){
			out.println("Teilhaber");
			out.indent();
			ArrayList gemTeilv=(ArrayList)gemeinschaftTeilhaberv.get(person.getobjectoid());
			for(int gemeinschaftTeilhaberi=0;gemeinschaftTeilhaberi<gemTeilv.size();gemeinschaftTeilhaberi++){
				IomObject gemeinschaftTeilhaber=(IomObject)gemTeilv.get(gemeinschaftTeilhaberi);
				IomObject refTeilhaber=gemeinschaftTeilhaber.getattrobj("Teilhaber",0);
				String teilhaberOid=refTeilhaber.getobjectrefoid();
				if(objv.containsKey(teilhaberOid)){
					IomObject teilhaber=(IomObject)objv.get(teilhaberOid);
					printPerson(teilhaber,objv,gemeinschaftTeilhaberv);
				}else{
					EhiLogger.logError("Objekt "+gemeinschaftTeilhaber.getobjectoid()+": enthlt Verweis auf nicht existierendes Objekt ("+teilhaberOid+")");
				}
				
			}
			out.unindent();
		}
		out.unindent();
	}
	
	/** konvertiert ein Grundsteuck.
	 * @param gs Grundstueck Objekt das konvertiert werden soll.
	 */
	private void printGS(IomObject gs) 
	{
		// Gemeinde
		IomObject gemeinde=gs.getattrobj("Gemeinde",0);
		if(gemeinde!=null){
			StringBuffer txt=new StringBuffer();
			txt.append("Gemeinde");
			String sep=" ";
			String Name=gemeinde.getattrvalue("Name");
			if(Name!=null){
				txt.append(sep+"Name "+Name);sep=", ";
			}
			String BFSNr=gemeinde.getattrvalue("BFSNr");
			if(BFSNr!=null){
				txt.append(sep+"BFSNr "+BFSNr);sep=", ";
			}
			out.println(txt);
		}
		// GBPlan
		int gbplanc=gs.getattrvaluecount("GBPlaene");
		for(int gbplani=0;gbplani<gbplanc;gbplani++){
			IomObject gbplan=gs.getattrobj("GBPlaene",gbplani);
			if(gbplan!=null){
				StringBuffer txt=new StringBuffer();
				txt.append("GBPlan");
				String sep=" ";
				String Nummer=gbplan.getattrvalue("Nummer");
				if(Nummer!=null){
					txt.append(sep+"Nummer "+Nummer);sep=", ";
				}
				String NBIdent=gbplan.getattrvalue("NBIdent");
				if(NBIdent!=null){
					txt.append(sep+"NBIdent "+NBIdent);sep=", ";
				}
				out.println(txt);
			}
		}
		// Flurname
		int flurnamec=gs.getattrvaluecount("Flurnamen");
		for(int flurnamei=0;flurnamei<flurnamec;flurnamei++){
			IomObject flurname=gs.getattrobj("Flurnamen",flurnamei);
			if(flurname!=null){
				StringBuffer txt=new StringBuffer();
				txt.append("Flurname");
				String sep=" ";
				String Name=flurname.getattrvalue("Name");
				if(Name!=null){
					txt.append(sep+Name);sep=", ";
				}
				out.println(txt);
			}
		}
		// Bodbd
		int bodbdc=gs.getattrvaluecount("BodbdAnteil");
		for(int bodbdi=0;bodbdi<bodbdc;bodbdi++){
			IomObject bodbd=gs.getattrobj("BodbdAnteil",bodbdi);
			if(bodbd!=null){
				StringBuffer txt=new StringBuffer();
				txt.append("Bodenbedeckung");
				String sep=" ";
				String Art=bodbd.getattrvalue("Art");
				if(Art!=null){
					txt.append(sep+"Art "+Art);sep=", ";
				}
				String Flaechenmass=bodbd.getattrvalue("Flaechenmass");
				if(Flaechenmass!=null){
					txt.append(sep+"Flaechenmass "+Flaechenmass);sep=", ";
				}
				String ArtBezeichnung=bodbd.getattrvalue("ArtBezeichnung");
				if(ArtBezeichnung!=null){
					txt.append(sep+"ArtBezeichnung "+ArtBezeichnung);sep=", ";
				}
				out.println(txt);
			}
		}
	}

	/** liefert alle Anteil Objekte zu einem Flaeche Objekt.
	 * @param flaecheOid oid des Objektes Flaeche
	 * @param objv alle Objekte map<String oid,IomObject>
	 * @return Liste der Anteil Objekte list<IomObject Anteil>
	 */
	private ArrayList getLiegtAufGS(String flaecheOid,HashMap objv,ArrayList assocv)
	{
		ArrayList ret=new ArrayList();
		Iterator oidi=assocv.iterator();
		while(oidi.hasNext()){
			IomObject iomObj=(IomObject)oidi.next();
			String aclass=iomObj.getobjecttag();
			// Ist es ein Assoziationsobjekt?
			if(aclass.equals(ILI_GSBESCHR+".Anteil")){
				IomObject refObj=iomObj.getattrobj("flaeche",0);
				if(refObj!=null){
					// Ist das ein Anteil zur gegebenen Flaeche?
					if(flaecheOid.equals(refObj.getobjectrefoid())){
						// In Liste aufnehmen
						ret.add(iomObj);
					}
				}
			}
		}
		return ret;
	}
	/** liefert die Liste der betroffenen Grundstuecke zu einer AVMutation
	 * @param mutOid oid des Objektes AVMutation
	 * @param objv alle Objekte map<String oid,IomObject>
	 * @return Liste der betroffenen Grundstuecke list<IomObject Grundstueck>
	 */
	private ArrayList getBetrfGS(String mutOid,HashMap objv,ArrayList assocv)
	{
		ArrayList ret=new ArrayList();
		Iterator oidi=assocv.iterator();
		while(oidi.hasNext()){
			IomObject iomObj=(IomObject)oidi.next();
			String aclass=iomObj.getobjecttag();
			// Ist es ein Assoziationsobjekt?
			if(aclass.equals(ILI_MUT+".AVMutationBetroffeneGrundstuecke")){
				IomObject refObj=iomObj.getattrobj("AVMutation",0);
				if(refObj!=null){
					// Ist das Grundstck Teil der Mutation?
					if(mutOid.equals(refObj.getobjectrefoid())){
						IomObject refGS=iomObj.getattrobj("betroffeneGrundstuecke",0);
						// In Liste aufnehmen
						String gsOid=refGS.getobjectrefoid();
						if(objv.containsKey(gsOid)){
							ret.add(objv.get(gsOid));
						}else{
							EhiLogger.logError("Objekt "+iomObj.getobjectoid()+": enthlt Verweis auf nicht existierendes Objekt ("+gsOid+")");
						}
					}
				}
			}
		}
		return ret;
	}

	/** formatiert eine GrundstueckNummer.
	 * @param gsNr Strukturelement GrundstueckNummer.
	 * @param tid OID des Objektes das das Strukturelement GrundstueckNummer enthaelt. Kann null sein.
	 * @return konvertierte Grundstuecknummer.
	 */
	private String formatGSNr(IomObject gsNr,String tid)
	{
		StringBuffer buf=new StringBuffer();
		String sep="";
		String egrid=gsNr.getattrvalue("EGRID");
		if(egrid!=null){
			buf.append(sep+"EGRID "+egrid);sep=", ";
		}
		String Nummer=gsNr.getattrvalue("Nummer");
		if(Nummer!=null){
			buf.append(sep+"Nummer "+Nummer);sep=", ";
		}
		String NummerZusatz=gsNr.getattrvalue("NummerZusatz");
		if(NummerZusatz!=null){
			buf.append(sep+"NummerZusatz "+NummerZusatz);sep=", ";
		}
		String Amt=gsNr.getattrvalue("Amt");
		if(Amt!=null){
			buf.append(sep+"Amt "+Amt);sep=", ";
		}
		String SubKreis=gsNr.getattrvalue("SubKreis");
		if(SubKreis!=null){
			buf.append(sep+"SubKreis "+SubKreis);sep=", ";
		}
		String Los=gsNr.getattrvalue("Los");
		if(Los!=null){
			buf.append(sep+"Los "+Los);sep=", ";
		}
		if(tid!=null){
			buf.append(" (TID "+tid+")");
		}
		return buf.toString();
	}
	/** formatiert eine Gebaeudenummer.
	 * Es gibt keine Gebaeudenummer Struktur. Diese Funktion verwendet alle Attribute aus der Klasse Gebaeude die als Nummer in Frage kommen.
	 * @param gsNr Gebaeude Objekt.
	 * @return fomatierte Gebaeudenummer.
	 */
	private String formatGebaeudeNr(IomObject gsNr)
	{
		StringBuffer buf=new StringBuffer();
		String sep="";
		String egid=gsNr.getattrvalue("GWR_EGID");
		if(egid!=null){
			buf.append(sep+"GWR_EGID "+egid);sep=", ";
		}
		String Nummer=gsNr.getattrvalue("Nummer");
		if(Nummer!=null){
			buf.append(sep+"Nummer "+Nummer);sep=", ";
		}
		buf.append(" (TID "+gsNr.getobjectoid()+")");
		return buf.toString();
	}
}
